// This Source Code Form is subject to the terms of the Mozilla Public
// License, v. 2.0.If a copy of the MPL was not distributed with this
// file, You can obtain one at https ://mozilla.org/MPL/2.0/.

#include "Direction.h"

#include "Quaternion.h"
#include "Vector3.h"

#include <math.h>

namespace LinearAlgebra {
template <typename T>
DirectionOf<T>::DirectionOf() {
  this->horizontal = AngleOf<T>();
  this->vertical = AngleOf<T>();
}

template <typename T>
DirectionOf<T>::DirectionOf(AngleOf<T> horizontal, AngleOf<T> vertical) {
  this->horizontal = horizontal;
  this->vertical = vertical;
  Normalize();
};

template <typename T>
const DirectionOf<T> DirectionOf<T>::forward =
    DirectionOf<T>(AngleOf<T>(), AngleOf<T>());
template <typename T>
const DirectionOf<T> DirectionOf<T>::back =
    DirectionOf<T>(AngleOf<T>::Degrees(180), AngleOf<T>());
template <typename T>
const DirectionOf<T> DirectionOf<T>::up =
    DirectionOf<T>(AngleOf<T>(), AngleOf<T>::Degrees(90));
template <typename T>
const DirectionOf<T> DirectionOf<T>::down =
    DirectionOf<T>(AngleOf<T>(), AngleOf<T>::Degrees(-90));
template <typename T>
const DirectionOf<T> DirectionOf<T>::left =
    DirectionOf<T>(AngleOf<T>::Degrees(-90), AngleOf<T>());
template <typename T>
const DirectionOf<T> DirectionOf<T>::right =
    DirectionOf<T>(AngleOf<T>::Degrees(90), AngleOf<T>());

template <typename T>
Vector3 DirectionOf<T>::ToVector3() const {
  Quaternion q = Quaternion::Euler(-this->vertical.InDegrees(),
                                   this->horizontal.InDegrees(), 0);
  Vector3 v = q * Vector3::forward;
  return v;
}

template <typename T>
DirectionOf<T> DirectionOf<T>::FromVector3(Vector3 vector) {
  DirectionOf<T> d;
  d.horizontal = AngleOf<T>::Atan2(
      vector.Right(),
      vector
          .Forward());  // AngleOf<T>::Radians(atan2f(v.Right(), v.Forward()));
  d.vertical =
      AngleOf<T>::Degrees(-90) -
      AngleOf<T>::Acos(
          vector.Up());  // AngleOf<T>::Radians(-(0.5f * pi) - acosf(v.Up()));
  d.Normalize();
  return d;
}

template <typename T>
DirectionOf<T> DirectionOf<T>::Degrees(float horizontal, float vertical) {
  return DirectionOf<T>(AngleOf<T>::Degrees(horizontal),
                        AngleOf<T>::Degrees(vertical));
}

template <typename T>
DirectionOf<T> DirectionOf<T>::Radians(float horizontal, float vertical) {
  return DirectionOf<T>(AngleOf<T>::Radians(horizontal),
                        AngleOf<T>::Radians(vertical));
}

template <typename T>
bool DirectionOf<T>::operator==(const DirectionOf<T> direction) const {
  return (this->horizontal == direction.horizontal) &&
         (this->vertical == direction.vertical);
}

template <typename T>
DirectionOf<T> DirectionOf<T>::operator-() const {
  DirectionOf<T> r = DirectionOf<T>(this->horizontal + AngleOf<T>::Degrees(180),
                                    -this->vertical);
  return r;
}

template <typename T>
void DirectionOf<T>::Normalize() {
  if (this->vertical > AngleOf<T>::Degrees(90) ||
      this->vertical < AngleOf<T>::Degrees(-90)) {
    this->horizontal += AngleOf<T>::Degrees(180);
    this->vertical = AngleOf<T>::Degrees(180) - this->vertical;
  }
}

template class LinearAlgebra::DirectionOf<float>;
template class LinearAlgebra::DirectionOf<signed short>;
}